from __future__ import annotations

import asyncio
import contextlib
import datetime
import json
import random
import socket
import sys
import time
import urllib.parse
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

import aiohttp

from .globals import config
from .xml import dict_to_xml, etree


@dataclass(frozen=True)
class RawRequest:
    method: str
    url: str
    params: Mapping[str, str] = field(default_factory=dict)
    # data can be large, so don't put it in the repr
    data: bytes | bytearray | memoryview | None = field(repr=False, default=None)
    # headers can contain authorisation details, so don't put them in the repr
    headers: Mapping[str, str] = field(repr=False, default_factory=dict)

    @contextlib.asynccontextmanager
    async def raw_execute(self) -> AsyncIterator[aiohttp.ClientResponse]:
        """Actually execute the request, with no extra fluff."""
        ctx = config.session.request(
            method=self.method,
            url=self.url,
            params=self.params,
            data=self.data,
            headers=self.headers,
            allow_redirects=False,
            timeout=aiohttp.ClientTimeout(
                connect=config.connect_timeout, sock_read=config.read_timeout
            ),
            # Figuring out that some requests break because aiohttp adds some headers
            # automatically was not fun
            skip_auto_headers={"Content-Type"},
        )
        if config.debug_mode:
            print(f"[boostedblob] Making request: {self}", file=sys.stderr)
            now = time.time()
        async with ctx as resp:
            if config.debug_mode:
                duration = time.time() - now
                print(
                    f"[boostedblob] Completed request, took {duration:.3f}s: {self}",
                    file=sys.stderr,
                )
            yield resp

    @classmethod
    def from_request(cls, request: Request) -> RawRequest:
        assert isinstance(request.data, (bytes, bytearray, memoryview)) or request.data is None
        return cls(
            method=request.method,
            url=request.url,
            params=request.params,
            data=request.data,
            headers=request.headers,
        )


@dataclass(frozen=True)
class Request:
    method: str
    url: str
    params: Mapping[str, str] = field(default_factory=dict)
    # data can be large, so don't put it in the repr
    data: dict[str, Any] | bytes | None = field(repr=False, default=None)
    # headers can contain authorisation details, so don't put them in the repr
    headers: Mapping[str, str] = field(repr=False, default_factory=dict)
    success_codes: Sequence[int] = field(repr=False, default_factory=lambda: (200,))
    retry_codes: Sequence[int] = field(
        repr=False, default_factory=lambda: (408, 429, 500, 502, 503, 504)
    )
    failure_exceptions: Mapping[int, Exception] = field(repr=False, default_factory=dict)
    auth: Callable[[Request], Awaitable[RawRequest]] | None = field(repr=False, default=None)

    @contextlib.asynccontextmanager
    async def execute(self) -> AsyncIterator[aiohttp.ClientResponse]:
        """Execute the request, retrying as appropriate.

        Example usage:
        ```
        async with request.execute() as resp:
            print(resp.status)
        ```

        """
        rreq: RawRequest | None = None
        last_auth_t = 0.0

        for attempt, backoff in enumerate(
            exponential_sleep_generator(
                initial=config.backoff_initial,
                maximum=config.backoff_max,
                jitter_fraction=config.backoff_jitter_fraction,
            )
        ):
            now = time.time()
            if (now - last_auth_t) > config.request_reauth_seconds:
                if self.auth is not None:
                    rreq = await self.auth(self)
                else:
                    rreq = RawRequest.from_request(self)
                last_auth_t = now
            assert rreq is not None

            async with contextlib.AsyncExitStack() as stack:
                try:
                    resp = await stack.enter_async_context(rreq.raw_execute())
                except aiohttp.ClientConnectionError as e:
                    if isinstance(e, aiohttp.ClientConnectorError):
                        # azure accounts have unique urls and it's hard to tell apart
                        # an invalid hostname from a network error
                        url = urllib.parse.urlparse(self.url)
                        hostname = url.hostname
                        if hostname and hostname.endswith(".blob.core.windows.net"):
                            if await _bad_hostname_check(hostname):
                                raise FileNotFoundError(hostname) from None
                    error = RequestFailure(reason=type(e).__name__ + ": " + str(e), request=self)
                else:
                    if resp.status in self.success_codes:
                        yield resp
                        return

                    assert resp.reason is not None
                    reason = repr(resp.reason)
                    error = RequestFailure(reason=reason, request=self, status=resp.status)
                    if resp.status not in self.retry_codes:
                        if resp.status in self.failure_exceptions:
                            raise self.failure_exceptions[resp.status]
                        # If we're actually going to raise, read the body
                        body = (await resp.read()).decode("utf-8")
                        reason += "\nBody:\n" + body + "\nHeaders:\n" + str(resp.headers)
                        raise RequestFailure(reason=reason, request=self, status=resp.status)
                    else:
                        if config.debug_mode:
                            body = (await resp.read()).decode("utf-8")
                            reason += "\nBody:\n" + body + "\nHeaders:\n" + str(resp.headers)
                        error = RequestFailure(reason=reason, request=self, status=resp.status)

            if attempt >= config.retry_limit:
                raise error

            if config.debug_mode or attempt + 1 >= 3:
                print(
                    f"[boostedblob] Error when executing request on attempt {attempt + 1}, "
                    f"sleeping for {backoff:.1f}s before retrying. Details:\n{error}",
                    file=sys.stderr,
                )
            await asyncio.sleep(backoff)

    async def execute_reponseless(self) -> None:
        """Helper to execute the request when we don't have further need of the response."""
        async with self.execute():
            pass


class RequestFailure(Exception):
    def __init__(self, reason: str, request: Request, status: Optional[int] = None):
        self.reason = reason
        self.request = request
        self.status = status

    def __str__(self) -> str:
        return f"Reason: {self.reason}\nRequest: {self.request}\nStatus: {self.status}"


async def execute_retrying_read(request: Request) -> bytes:
    # Retrying aiohttp.ServerTimeoutError is pretty straightforward
    # ClientPayloadError might be an aiohttp bug, see:
    # https://github.com/aio-libs/aiohttp/issues/4581
    # https://github.com/aio-libs/aiohttp/issues/3904#issuecomment-737094416
    for attempt, backoff in enumerate(
        exponential_sleep_generator(
            initial=config.backoff_initial,
            maximum=config.backoff_max,
            jitter_fraction=config.backoff_jitter_fraction,
        )
    ):
        try:
            async with request.execute() as resp:
                return await resp.read()
        except (aiohttp.ServerTimeoutError, aiohttp.ClientPayloadError) as error:
            if attempt >= config.retry_limit:
                raise

            if config.debug_mode or attempt + 1 >= 3:
                print(
                    f"[boostedblob] Error when executing request on attempt {attempt + 1}, "
                    f"sleeping for {backoff:.1f}s before retrying. Details:\n{error}",
                    file=sys.stderr,
                )
            await asyncio.sleep(backoff)
    raise AssertionError


# ==============================
# cloud specific utilities
# ==============================


async def azure_auth_req(request: Request, *, auth: Optional[Tuple[str, str]] = None) -> RawRequest:
    """Return a Request that can be submitted to Azure Blob."""
    u = urllib.parse.urlparse(request.url)
    account = u.netloc.split(".")[0]
    parts = u.path.split("/", maxsplit=2)
    container = parts[1] if len(parts) >= 2 else None
    if auth is None:
        auth = await config.azure_access_token_manager.get_token(key=(account, container))
    assert auth is not None
    kind, token = auth

    headers = dict(request.headers)
    # https://docs.microsoft.com/en-us/rest/api/storageservices/previous-azure-storage-service-versions
    headers["x-ms-version"] = "2023-05-03"
    headers["x-ms-date"] = datetime.datetime.now(datetime.timezone.utc).strftime(
        "%a, %d %b %Y %H:%M:%S GMT"
    )

    from .azure_auth import OAUTH_TOKEN, SHARED_KEY, sign_request_with_shared_key

    data = request.data
    if data is not None and not isinstance(data, (bytes, bytearray, memoryview)):
        data = dict_to_xml(data)

    result = RawRequest(
        method=request.method, url=request.url, params=request.params, headers=headers, data=data
    )

    # mutate headers to add the authorization header to result
    # the control flow is a little complex in that we pass result into sign_request_with_shared_key
    if kind == SHARED_KEY:
        # make sure we are signing a request that includes changes made in this function
        # i.e., the request has the x-ms-* headers added and data bytes-ified.
        assert "x-ms-date" in headers
        headers["Authorization"] = sign_request_with_shared_key(result, token)
    elif kind == OAUTH_TOKEN:
        headers["Authorization"] = f"Bearer {token}"

    return result


async def google_auth_req(request: Request, *, access_token: Optional[str] = None) -> RawRequest:
    """Return a Request that can be submitted to GCS."""
    if access_token is None:
        access_token = await config.google_access_token_manager.get_token(key="")

    headers = dict(request.headers)
    headers["Authorization"] = f"Bearer {access_token}"

    data = request.data
    if data is not None and not isinstance(data, (bytes, bytearray, memoryview)):
        data = json.dumps(data).encode("utf8")

    return RawRequest(
        method=request.method, url=request.url, params=request.params, headers=headers, data=data
    )


async def xml_page_iterator(request: Request) -> AsyncIterator[etree.Element]:
    params = dict(request.params)
    while True:
        request = Request(
            method=request.method,
            url=request.url,
            params=params,
            data=request.data,
            headers=request.headers,
            success_codes=request.success_codes,
            retry_codes=request.retry_codes,
            failure_exceptions=request.failure_exceptions,
            auth=request.auth,
        )
        body = await execute_retrying_read(request)

        result = etree.fromstring(body)
        yield result

        next_marker = result.findtext("NextMarker")
        if not next_marker:
            break
        params["marker"] = next_marker


async def json_token_page_iterator(request: Request) -> AsyncIterator[Dict[str, Any]]:
    params = dict(request.params)
    while True:
        request = Request(
            method=request.method,
            url=request.url,
            params=params,
            headers=request.headers,
            success_codes=(200, 404),
            retry_codes=request.retry_codes,
            failure_exceptions=request.failure_exceptions,
            auth=request.auth,
        )
        async with request.execute() as resp:
            if resp.status == 404:
                return
            result = await resp.json()
        yield result
        if "nextPageToken" not in result:
            break
        params["pageToken"] = result["nextPageToken"]


# ==============================
# helpers
# ==============================


def exponential_sleep_generator(
    initial: float, maximum: float, jitter_fraction: float, multiplier: float = 2
) -> Iterator[float]:
    """Yields amounts to sleep by.

    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    Full jitter is ``jitter_fraction == 1``
    Equal jitter is ``jitter_fraction == 0.5``
    No jitter is ``jitter_fraction == 0``
    """

    base = initial
    while True:
        sleep = base * (1 - jitter_fraction) + base * random.random() * jitter_fraction
        yield sleep
        base *= multiplier
        base = min(base, maximum)


_hostname_check_cache: Dict[str, Tuple[float, bool]] = {}


async def _bad_hostname_check(hostname: str) -> bool:
    """Return True if the hostname likely does not exist.

    If the hostname exists, or there is uncertainty, return False.
    """

    async def inner(hostname: str) -> bool:
        loop = asyncio.get_running_loop()
        try:
            await loop.getaddrinfo(hostname, None, family=socket.AF_INET)
            # no errors encountered, the hostname exists
            return False
        except OSError as e:
            if e.errno != socket.EAI_NONAME:
                # we got some sort of other socket error, so it's unclear if the host exists or not
                return False

            # On linux, most temporary name resolution failures return EAI_AGAIN, however it
            # appears that some EAI_NONAME can be triggered by SERVFAIL, instead of just NXDOMAIN
            # On other platforms, it's unclear how to differentiate between a temporary failure
            # and a permanent one.
            # Either way, assume that if we can lookup bing.com, then it's likely the hostname
            # does not exist
            try:
                await loop.getaddrinfo("www.bing.com", None, family=socket.AF_INET)
            except OSError:
                # if we can't resolve bing, then the network is likely down and we don't know if
                # the hostname exists or not
                return False
            # in this case, we could resolve bing, but not the original hostname. likely the
            # hostname does not exist (though this is definitely not a foolproof check)
            return True

    # maybe this cache is a little bit overkill...
    now = time.time()
    if hostname in _hostname_check_cache and _hostname_check_cache[hostname][0] >= now:
        return _hostname_check_cache[hostname][1]
    ret = await inner(hostname)
    _hostname_check_cache[hostname] = (now + 10, ret)
    return ret



from __future__ import annotations

import configparser
import os
import re
from typing import Any, AsyncIterator, Iterator, Mapping, NamedTuple, Optional, Union

from . import google_auth
from .path import (
    AzurePath,
    AzureStat,
    BasePath,
    BlobPath,
    CloudPath,
    GooglePath,
    GoogleStat,
    LocalPath,
    LocalStat,
    Stat,
    isfile,
    pathdispatch,
)
from .request import (
    Request,
    azure_auth_req,
    google_auth_req,
    json_token_page_iterator,
    xml_page_iterator,
)
from .xml import etree

# ==============================
# DirEntry
# ==============================


class DirEntry(NamedTuple):
    path: BasePath
    is_dir: bool
    is_file: bool
    stat: Optional[Stat]

    @staticmethod
    def from_dirpath(path: BasePath) -> DirEntry:
        assert path.is_directory_like()
        return DirEntry(path=path, is_dir=True, is_file=False, stat=None)

    @staticmethod
    def from_path_stat(path: BasePath, stat: Stat) -> DirEntry:
        assert not path.is_directory_like()
        return DirEntry(path=path, is_dir=False, is_file=True, stat=stat)


# ==============================
# list_blobs
# ==============================


@pathdispatch
def list_blobs(
    path: Union[CloudPath, str], delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    """List all blobs whose prefix matches ``path``.

    :param path: The prefix we want to list.
    :param delimiter: Group blobs matching the prefix up to a delimiter. This allows us to emulate
        directories on blob storage.
    :param allow_prefix: Whether to allow prefixes that do not end with a delimiter (or "/").

    """
    raise ValueError(f"Unsupported path: {path}")


@list_blobs.register  # type: ignore
async def _azure_list_blobs(
    path: AzurePath, delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    prefix = path
    if not allow_prefix:
        _delimiter = delimiter or "/"
        assert not prefix.blob or prefix.blob.endswith(_delimiter)
    if not path.container:
        if path.blob:
            raise ValueError("Path is missing container; do you have an extra slash?")
        if delimiter != "/":
            raise ValueError("Cannot list blobs in storage account; must specify container")
        async for entry in _azure_list_containers(prefix.account):
            yield entry
        return

    params = {}
    if delimiter is not None:
        params["delimiter"] = delimiter
    it = xml_page_iterator(
        Request(
            method="GET",
            url=prefix.format_url("https://{account}.blob.core.windows.net/{container}"),
            params=dict(comp="list", restype="container", prefix=prefix.blob, **params),
            auth=azure_auth_req,
        )
    )

    async for result in it:
        for entry in _azure_get_entries(prefix.account, prefix.container, result):
            yield entry


@list_blobs.register  # type: ignore
async def _google_list_blobs(
    path: GooglePath, delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    prefix = path
    if not allow_prefix:
        _delimiter = delimiter or "/"
        assert not prefix.blob or prefix.blob.endswith(_delimiter)
    if not path.bucket:
        if delimiter != "/":
            raise ValueError("Cannot list blobs across buckets")
        async for entry in _google_list_buckets():
            yield entry
        return

    params = {}
    if delimiter is not None:
        params["delimiter"] = delimiter
    it = json_token_page_iterator(
        Request(
            method="GET",
            url=prefix.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o"),
            params=dict(prefix=prefix.blob, **params),
            auth=google_auth_req,
        )
    )

    async for result in it:
        for entry in _google_get_entries(prefix.bucket, result):
            yield entry


# ==============================
# scandir
# ==============================


@pathdispatch
def scandir(path: Union[BasePath, BlobPath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over the entries in a directory.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@scandir.register  # type: ignore
async def _cloud_scandir(path: CloudPath) -> AsyncIterator[DirEntry]:
    dirpath = path.ensure_directory_like()
    subpath_exists = False
    async for entry in list_blobs(dirpath, delimiter="/"):
        subpath_exists = True
        if entry.path == dirpath:
            continue
        yield entry

    # If we find nothing, then run some checks so we throw the appropriate error.
    # Doing this means we avoid extra requests in the happy path.
    if not subpath_exists:
        if isinstance(path, AzurePath) and not path.blob:
            # Special case empty containers (at this point we're guaranteed the container exists)
            # TODO: double check this when anonymous containers work
            return
        if not await isfile(path):
            raise FileNotFoundError(path)
        raise NotADirectoryError(path)


@scandir.register  # type: ignore
async def _local_scandir(path: LocalPath) -> AsyncIterator[DirEntry]:
    path_absolute = os.path.abspath(path)
    for entry in os.scandir(path_absolute):
        # entry.path is absolute since we're passing an absolute path to os.scandir
        entry_path = LocalPath(entry.path)
        if entry.is_dir():
            yield DirEntry(
                path=entry_path.ensure_directory_like(), is_dir=True, is_file=False, stat=None
            )
        else:
            yield DirEntry(path=entry_path, is_dir=False, is_file=True, stat=_os_safe_stat(entry))


# ==============================
# listdir
# ==============================


@pathdispatch
def listdir(path: Union[BasePath, BlobPath, str]) -> AsyncIterator[BasePath]:
    """Iterate over the names of entries in a directory.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@listdir.register  # type: ignore
async def _cloud_listdir(path: CloudPath) -> AsyncIterator[CloudPath]:
    async for entry in scandir(path):
        assert isinstance(entry.path, type(path))
        yield entry.path


@listdir.register  # type: ignore
async def _local_listdir(path: LocalPath) -> AsyncIterator[LocalPath]:
    for entry in os.scandir(path):
        entry_path = LocalPath(os.path.normpath(entry.path))
        yield entry_path.ensure_directory_like() if entry.is_dir() else entry_path


# ==============================
# scantree
# ==============================


@pathdispatch
def scantree(path: Union[BasePath, BlobPath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over file entries in the directory tree rooted at path.

    :param path: The root of the tree we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@scantree.register  # type: ignore
async def _cloud_scantree(path: CloudPath) -> AsyncIterator[DirEntry]:
    dirpath = path.ensure_directory_like()
    subpath_exists = False
    async for entry in list_blobs(dirpath, None):
        subpath_exists = True
        if entry.path == dirpath:
            continue
        yield entry

    # If we find nothing, then run some checks so we throw the appropriate error.
    # Doing this means we avoid extra requests in the happy path.
    if not subpath_exists:
        if not await isfile(path):
            raise FileNotFoundError(path)
        raise NotADirectoryError(path)


@scantree.register  # type: ignore
async def _local_scantree(path: LocalPath) -> AsyncIterator[DirEntry]:
    path_absolute = os.path.abspath(path)

    def inner(current: str) -> Iterator[DirEntry]:
        entries = list(os.scandir(current))
        for entry in entries:
            if entry.is_dir():
                yield from inner(entry.path)
            else:
                yield DirEntry(
                    path=LocalPath(entry.path),
                    is_dir=False,
                    is_file=True,
                    stat=_os_safe_stat(entry),
                )

    for entry in inner(path_absolute):
        yield entry


# ==============================
# listtree
# ==============================


@pathdispatch
def listtree(path: Union[BasePath, BlobPath, str]) -> AsyncIterator[BasePath]:
    """List files in the directory tree rooted at path.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@listtree.register  # type: ignore
async def _cloud_listtree(path: CloudPath) -> AsyncIterator[CloudPath]:
    async for entry in scantree(path):
        assert isinstance(entry.path, type(path))
        yield entry.path


@listtree.register  # type: ignore
async def _local_listtree(path: LocalPath) -> AsyncIterator[LocalPath]:
    for base, _subdirs, files in os.walk(path):
        for name in files:
            yield LocalPath(base) / name


# ==============================
# glob_scandir
# ==============================


@pathdispatch
def glob_scandir(path: Union[BasePath, BlobPath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over a glob in a directory.

    :param path: The glob to match against.

    """
    raise ValueError(f"Unsupported path: {path}")


@glob_scandir.register  # type: ignore
async def _azure_glob_scandir(path: AzurePath) -> AsyncIterator[DirEntry]:
    if "*" in path.account:
        raise ValueError("Cannot use wildcard in storage account")
    if "*" in path.container:
        if path.blob:
            raise ValueError("Currently only supports wildcards inside the filename")
        pattern = _glob_to_regex(path.container)
        async for entry in _azure_list_containers(path.account):
            assert isinstance(entry.path, type(path))
            if re.match(pattern, entry.path.container):
                yield entry
        return
    if "*" in os.path.dirname(path.blob):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    blob_prefix = path.blob.split("*", maxsplit=1)[0]
    prefix = AzurePath(account=path.account, container=path.container, blob=blob_prefix)
    async for entry in list_blobs(prefix, delimiter="/", allow_prefix=True):
        assert isinstance(entry.path, type(path))
        if entry.path == path.parent.ensure_directory_like():
            # skip directory file marker of the parent, if present
            continue
        if re.match(pattern, entry.path.name):
            yield entry


@glob_scandir.register  # type: ignore
async def _google_glob_scandir(path: GooglePath) -> AsyncIterator[DirEntry]:
    if "*" in path.bucket:
        if path.blob:
            raise ValueError("Currently only supports wildcards inside the filename")
        pattern = _glob_to_regex(path.bucket)
        async for entry in _google_list_buckets():
            assert isinstance(entry.path, type(path))
            if re.match(pattern, entry.path.bucket):
                yield entry
        return
    if "*" in os.path.dirname(path.blob):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    blob_prefix = path.blob.split("*", maxsplit=1)[0]
    prefix = GooglePath(bucket=path.bucket, blob=blob_prefix)
    async for entry in list_blobs(prefix, delimiter="/", allow_prefix=True):
        assert isinstance(entry.path, type(path))
        if entry.path == path.parent.ensure_directory_like():
            # skip directory file marker of the parent, if present
            continue
        if re.match(pattern, entry.path.name):
            yield entry


@glob_scandir.register  # type: ignore
async def _local_glob_scandir(path: LocalPath) -> AsyncIterator[DirEntry]:
    if "*" in os.path.dirname(path):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    async for entry in scandir(path.parent):
        if re.match(pattern, entry.path.name):
            yield entry


# ==============================
# helpers
# ==============================


def _azure_get_entries(account: str, container: str, result: etree.Element) -> Iterator[DirEntry]:
    blobs = result.find("Blobs")
    assert blobs is not None
    blob: str
    for bp in blobs.iterfind("BlobPrefix"):
        blob = bp.findtext("Name")  # type: ignore
        path = AzurePath(account, container, blob)
        yield DirEntry.from_dirpath(path)
    for b in blobs.iterfind("Blob"):
        blob = b.findtext("Name")  # type: ignore
        path = AzurePath(account, container, blob)
        if blob.endswith("/"):
            yield DirEntry.from_dirpath(path)
        else:
            # this is much more efficient than doing repeated finds, even if we get extra fields
            props = {el.tag: el.text for el in b.find("Properties")}  # type: ignore
            yield DirEntry.from_path_stat(path, AzureStat(props))


def _google_get_entries(bucket: str, result: Mapping[str, Any]) -> Iterator[DirEntry]:
    if "prefixes" in result:
        for blob in result["prefixes"]:
            path = GooglePath(bucket, blob)
            yield DirEntry.from_dirpath(path)
    if "items" in result:
        for item in result["items"]:
            blob = item["name"]
            path = GooglePath(bucket, blob)
            if item["name"].endswith("/"):
                yield DirEntry.from_dirpath(path)
            else:
                yield DirEntry.from_path_stat(path, GoogleStat(item))


async def _azure_list_containers(account: str) -> AsyncIterator[DirEntry]:
    it = xml_page_iterator(
        Request(
            method="GET",
            url=f"https://{account}.blob.core.windows.net/",
            params=dict(comp="list"),
            failure_exceptions={404: FileNotFoundError(AzurePath(account, "", ""))},
            auth=azure_auth_req,
        )
    )
    async for result in it:
        containers = result.find("Containers")
        assert containers is not None
        if len(containers) == 0:
            raise ValueError(f"No containers found in storage account {account}")

        for container in containers.iterfind("Container"):
            name: str = container.find("Name").text  # type: ignore
            yield DirEntry(path=AzurePath(account, name, ""), is_dir=True, is_file=False, stat=None)


async def _google_list_buckets(project: Optional[str] = None) -> AsyncIterator[DirEntry]:
    if project is None:
        try:
            gconfig = configparser.ConfigParser()
            gconfig.read(
                os.path.join(google_auth.default_gcloud_path(), "configurations/config_default")
            )
            project = gconfig["core"]["project"]
        except (KeyError, FileNotFoundError):
            raise ValueError(
                "Could not determine project in which to list buckets; try setting it "
                "with `gsutil config`"
            ) from None

    it = json_token_page_iterator(
        Request(
            method="GET",
            url="https://storage.googleapis.com/storage/v1/b",
            params=dict(project=project),
            auth=google_auth_req,
        )
    )
    async for result in it:
        buckets = result["items"]
        assert isinstance(buckets, list)
        for bucket in buckets:
            yield DirEntry(
                path=GooglePath(bucket["name"], ""), is_dir=True, is_file=False, stat=None
            )


def _glob_to_regex(pattern: str) -> str:
    tokens = (token for token in re.split("([*]+)", pattern) if token != "")
    regexp = ""
    for token in tokens:
        if token == "*":
            regexp += r"[^/]*"
        elif token == "**":
            regexp += r".*"
        else:
            regexp += re.escape(token)
    return regexp + r"/?$"


def _os_safe_stat(entry: os.DirEntry[Any]) -> Optional[LocalStat]:
    # stat can fail and we don't want that to affect listing operations
    # In particular, we ran into the equivalent of the following raising FileNotFoundError:
    # next(x for x in os.scandir("/") if x.name == ".VolumeIcon.icns").stat()
    try:
        return LocalStat(entry.stat())
    except Exception:
        return None



from __future__ import annotations

import asyncio
import contextlib
import functools
import json
import os
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
)

import aiohttp

from . import azure_auth, google_auth

MB = 2**20


T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])


class TokenManager(Generic[T]):
    """Automatically refresh a token when it expires."""

    def __init__(self, get_token_fn: Callable[[T], Awaitable[Tuple[Any, float]]]) -> None:
        self._get_token_fn = get_token_fn
        self._tokens: Dict[T, Any] = {}
        self._expirations: Dict[T, float] = {}
        self._locks: Dict[asyncio.AbstractEventLoop, asyncio.Lock] = {}

    async def get_token(self, key: T) -> Any:
        if not self._tokens:
            self.load_state()

        running_loop = asyncio.get_running_loop()
        if running_loop not in self._locks:
            self._locks[running_loop] = asyncio.Lock()
        lock = self._locks[running_loop]

        now = time.time()
        expiration = self._expirations.get(key)

        if expiration is None or (now + config.token_early_expiration_seconds) > expiration:
            async with lock:
                refresh_expiration = self._expirations.get(key)
                if refresh_expiration == expiration:
                    self._tokens[key], self._expirations[key] = await self._get_token_fn(key)
                    assert self._expirations[key] is not None
                    self.dump_state()

        assert key in self._tokens
        return self._tokens[key]

    def get_cache_file(self) -> str:
        qn = self._get_token_fn.__module__ + "." + self._get_token_fn.__qualname__
        return os.path.expanduser(f"~/.config/bbb/{qn}.json")

    def load_state(self) -> None:
        if os.environ.get("BBB_DISABLE_CACHE"):
            return
        try:
            cache_file = self.get_cache_file()
            with open(cache_file) as f:
                state = json.load(f)
            if state.get("__version__") != 1:
                os.remove(cache_file)
                raise RuntimeError("Outdated format")
        except Exception as e:
            if config.debug_mode:
                print(f"[boostedblob] Error while loading token cache: {e}", file=sys.stderr)
            return

        for t in state["tokens"]:
            key: Any = t["key"]
            key = tuple(key) if isinstance(key, list) else key
            self._tokens[key] = t["token"]
            self._expirations[key] = t["expiration"]

    def dump_state(self) -> None:
        if os.environ.get("BBB_DISABLE_CACHE"):
            return
        now = time.time()
        state = {
            "__version__": 1,
            "tokens": [
                {
                    "key": k,
                    "token": self._tokens[k],
                    # lie about token expiry, so there's a limit on how long bbb will used a
                    # cached token
                    "expiration": min(exp, now + config.token_early_expiration_seconds * 2),
                }
                for k, exp in self._expirations.items()
                if exp > now  # only save tokens that have not expired
            ],
        }
        try:
            cache_file = self.get_cache_file()
            os.makedirs(os.path.dirname(cache_file), exist_ok=True)
            with open(cache_file, "w") as f:
                json.dump(state, f)
        except Exception as e:
            if config.debug_mode:
                print(f"[boostedblob] Error while dumping token cache: {e}", file=sys.stderr)


@dataclass
class Config:
    # Avoid the default executor because https://bugs.python.org/issue35279
    executor: ThreadPoolExecutor = field(default_factory=lambda: ThreadPoolExecutor(max_workers=32))

    debug_mode: bool = bool(os.environ.get("BBB_DEBUG"))
    storage_account_key_fallback: bool = bool(os.environ.get("BBB_SA_KEY_FALLBACK"))

    chunk_size: int = 16 * MB

    connect_timeout: float = 20.0
    read_timeout: float = 60.0
    backoff_initial: float = 0.1
    backoff_max: float = 60.0
    backoff_jitter_fraction: float = 0.9
    retry_limit: int = 25

    token_early_expiration_seconds: int = 300
    request_reauth_seconds: int = 300

    azure_access_token_manager: TokenManager[Tuple[str, Optional[str]]] = field(
        default_factory=lambda: TokenManager(azure_auth.get_access_token)
    )
    azure_sas_token_manager: TokenManager[Tuple[str, Optional[str]]] = field(
        default_factory=lambda: TokenManager(azure_auth.get_sas_token)
    )
    google_access_token_manager: TokenManager[str] = field(
        default_factory=lambda: TokenManager(google_auth.get_access_token)
    )

    _sessions: Dict[asyncio.AbstractEventLoop, aiohttp.ClientSession] = field(
        default_factory=dict, init=False
    )

    def _get_session(self) -> Optional[aiohttp.ClientSession]:
        return self._sessions.get(asyncio.get_running_loop())

    def _set_session(self, session: Optional[aiohttp.ClientSession]) -> None:
        running_loop = asyncio.get_running_loop()
        if session is None:
            self._sessions.pop(running_loop, None)
            return
        # In case we're doing repeated asyncio.run, clean up old sessions
        for l in list(self._sessions):
            if l.is_closed():
                del self._sessions[l]
        if hasattr(session, "_loop"):  # hasattr in case aiohttp changes
            assert session._loop is running_loop
        self._sessions[running_loop] = session

    async def _close_session(self) -> None:
        session = self._sessions.pop(asyncio.get_running_loop(), None)
        if session:
            await session.close()

    @property
    def session(self) -> aiohttp.ClientSession:
        loop = asyncio.get_running_loop()
        try:
            return self._sessions[loop]
        except KeyError:
            set_event_loop_exception_handler()  # there could be a better place for this
            session = _create_session()
            assert session._loop is loop
            self._set_session(session)
            return session

    @session.setter
    def session(self, session: aiohttp.ClientSession) -> None:
        import warnings

        warnings.warn(
            "Setting config.session is deprecated. Use config.configure(session=...) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._set_session(session)


config: Config = Config()


@contextlib.contextmanager
def configure(**kwargs: Any) -> Iterator[None]:
    old_session = None
    if "session" in kwargs:
        old_session = config._get_session()
        config._set_session(kwargs.pop("session"))

    original = {k: getattr(config, k) for k in kwargs}
    config.__dict__.update(**kwargs)
    try:
        yield
    finally:
        config.__dict__.update(**original)
        config._set_session(old_session)


def _create_session() -> aiohttp.ClientSession:
    # NB: We currently seem to leak file descriptors. E.g., if you run with -X dev, you'll see
    # "ResourceWarning: unclosed resource <TCPTransport ...>"
    # This can be fixed by changing the following line to:
    # connector = aiohttp.TCPConnector(limit=0, force_close=True, enable_cleanup_closed=True)
    # Both additional arguments are apparently necessary. Unfortunately this negates the
    # benefits of reusing connections. If you only have a single aiohttp.ClientSession, as is
    # recommended, this isn't really a problem.
    # Also note:
    # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
    # https://github.com/aio-libs/aiohttp/issues/1925#issuecomment-715977247
    # While the sleep suggested doesn't work, it does indicate that this is a problem for
    # aiohttp in general.
    connector = aiohttp.TCPConnector(limit=1024, ttl_dns_cache=60)
    return aiohttp.ClientSession(connector=connector)


@contextlib.asynccontextmanager
async def session_context() -> AsyncIterator[None]:
    preexisting_session = bool(config._get_session())
    try:
        yield
    finally:
        if not preexisting_session:
            await config._close_session()


def ensure_session(fn: F) -> F:
    @functools.wraps(fn)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        async with session_context():
            return await fn(*args, **kwargs)

    return wrapper  # type: ignore[return-value]


def set_event_loop_exception_handler() -> None:
    loop = asyncio.get_running_loop()

    def handler(
        loop: asyncio.AbstractEventLoop, context: Dict[str, Any]
    ) -> None:  # pragma: no cover
        message = context["message"]
        exception = context.get("exception", Exception)

        print(f"ERROR (from event loop): {type(exception).__name__}: {message}", file=sys.stderr)
        if loop.get_debug():
            loop.default_exception_handler(context)

    loop.set_exception_handler(handler)
